@@ -27,8 +27,9 @@ use crate::classic::clvm::sexp::{enlist, proper_list, sexp_as_bin};
 use crate::classic::clvm::KEYWORD_FROM_ATOM;
 use crate::classic::clvm_tools::binutils::{assemble_from_ir, disassemble, disassemble_with_kw};
 use crate::classic::clvm_tools::clvmc::detect_modern;
-use crate::classic::clvm_tools::debug::trace_pre_eval;
-use crate::classic::clvm_tools::debug::{trace_to_table, trace_to_text};
+use crate::classic::clvm_tools::debug::{
+    check_unused, trace_pre_eval, trace_to_table, trace_to_text,
+};
 use crate::classic::clvm_tools::ir::reader::read_ir;
 use crate::classic::clvm_tools::sha256tree::sha256tree;
 use crate::classic::clvm_tools::stages;
@@ -117,7 +118,7 @@ pub fn call_tool<'a>(
             args = a;
         }
         Err(e) => {
-            println!("{:?}", e);
+            println!("{}", e);
             return;
         }
     }
@@ -237,13 +238,17 @@ impl ArgumentValueConv for StageImport {
 pub fn run(args: &Vec<String>) {
     let mut s = Stream::new(None);
     launch_tool(&mut s, args, &"run".to_string(), 2);
-    io::stdout().write_all(s.get_value().data());
+    io::stdout()
+        .write_all(s.get_value().data())
+        .expect("stdout");
 }
 
 pub fn brun(args: &Vec<String>) {
     let mut s = Stream::new(None);
     launch_tool(&mut s, args, &"brun".to_string(), 0);
-    io::stdout().write_all(s.get_value().data());
+    io::stdout()
+        .write_all(s.get_value().data())
+        .expect("stdout");
 }
 
 pub fn cldb(args: &Vec<String>) {
@@ -253,6 +258,7 @@ pub fn cldb(args: &Vec<String>) {
         prog: format!("clvm_tools {}", tool_name),
     };
 
+    let mut search_paths = Vec::new();
     let mut parser = ArgumentParser::new(Some(props));
     parser.add_argument(
         vec!["-i".to_string(), "--include".to_string()],
@@ -318,6 +324,15 @@ pub fn cldb(args: &Vec<String>) {
         }
     }
 
+    if let Some(ArgumentValue::ArgArray(v)) = parsedArgs.get("include") {
+        for p in v {
+            match p {
+                ArgumentValue::ArgString(_, s) => search_paths.push(s.to_string()),
+                _ => {}
+            }
+        }
+    }
+
     match parsedArgs.get("path_or_code") {
         Some(ArgumentValue::ArgString(file, path_or_code)) => {
             input_file = file.clone();
@@ -357,7 +372,9 @@ pub fn cldb(args: &Vec<String>) {
     let use_filename = input_file
         .clone()
         .unwrap_or_else(|| "*command*".to_string());
-    let opts = Rc::new(DefaultCompilerOpts::new(&use_filename)).set_optimize(do_optimize);
+    let opts = Rc::new(DefaultCompilerOpts::new(&use_filename))
+        .set_optimize(do_optimize)
+        .set_search_paths(&search_paths);
 
     let mut use_symbol_table = symbol_table.unwrap_or_else(|| HashMap::new());
     let unopt_res = compile_file(
@@ -567,6 +584,13 @@ pub fn launch_tool(
     };
 
     let mut parser = ArgumentParser::new(Some(props));
+    parser.add_argument(
+        vec!["-s".to_string(), "--stage".to_string()],
+        Argument::new()
+            .set_type(Rc::new(StageImport {}))
+            .set_help("stage number to include".to_string())
+            .set_default(ArgumentValue::ArgInt(default_stage as i64)),
+    );
     parser.add_argument(
         vec!["--strict".to_string()],
         Argument::new()
@@ -579,13 +603,6 @@ pub fn launch_tool(
             .set_action(TArgOptionAction::StoreTrue)
             .set_help("Read program and environment as hexadecimal bytecode".to_string()),
     );
-    parser.add_argument(
-        vec!["-s".to_string(), "--stage".to_string()],
-        Argument::new()
-            .set_type(Rc::new(StageImport {}))
-            .set_help("stage number to include".to_string())
-            .set_default(ArgumentValue::ArgInt(default_stage as i64)),
-    );
     parser.add_argument(
         vec!["-v".to_string(), "--verbose".to_string()],
         Argument::new()
@@ -669,6 +686,17 @@ pub fn launch_tool(
             .set_help("run optimizer".to_string()),
     );
 
+    if tool_name == "run" {
+        parser.add_argument(
+            vec!["--check-unused-args".to_string()],
+            Argument::new()
+                .set_action(TArgOptionAction::StoreTrue)
+                .set_help(
+                    "check for unused uncurried parameters (by convention lower case)".to_string(),
+                ),
+        );
+    }
+
     let arg_vec = args[1..].to_vec();
     let parsedArgs: HashMap<String, ArgumentValue>;
 
@@ -691,6 +719,7 @@ pub fn launch_tool(
 
     let dpr;
     let run_program: Rc<dyn TRunProgram>;
+    let mut search_paths = Vec::new();
     match parsedArgs.get("include") {
         Some(ArgumentValue::ArgArray(v)) => {
             let mut bare_paths = Vec::with_capacity(v.len());
@@ -701,6 +730,7 @@ pub fn launch_tool(
                 }
             }
             let special_runner = run_program_for_search_paths(&bare_paths);
+            search_paths = bare_paths;
             dpr = special_runner.clone();
             run_program = special_runner;
         }
@@ -828,8 +858,46 @@ pub fn launch_tool(
         _ => {}
     }
 
+    // Add unused check.
+    let do_check_unused = parsedArgs
+        .get("check_unused_args")
+        .map(|a| match a {
+            ArgumentValue::ArgBool(true) => true,
+            _ => false,
+        })
+        .unwrap_or_else(|| false);
+
+    let dialect = input_sexp.and_then(|i| detect_modern(&mut allocator, i));
+    let mut stderr_output = |s| {
+        if let Some(_) = dialect {
+            eprintln!("{}", s);
+        } else {
+            stdout.write_string(s);
+        }
+    };
+
+    if do_check_unused {
+        let use_filename = input_file
+            .as_ref()
+            .map(|x| x.clone())
+            .unwrap_or_else(|| "*command*".to_string());
+        let opts = Rc::new(DefaultCompilerOpts::new(&use_filename)).set_search_paths(&search_paths);
+        match check_unused(opts, &input_program) {
+            Ok((success, output)) => {
+                stderr_output(output);
+                if !success {
+                    return;
+                }
+            }
+            Err(e) => {
+                stderr_output(format!("{}: {}\n", e.0.to_string(), e.1));
+                return;
+            }
+        }
+    }
+
     // In testing: short circuit for modern compilation.
-    if let Some(dialect) = input_sexp.and_then(|i| detect_modern(&mut allocator, i)) {
+    if let Some(dialect) = dialect {
         let do_optimize = parsedArgs
             .get("optimize")
             .map(|x| match x {
@@ -841,6 +909,7 @@ pub fn launch_tool(
         let use_filename = input_file.unwrap_or_else(|| "*command*".to_string());
         let opts = Rc::new(DefaultCompilerOpts::new(&use_filename))
             .set_optimize(do_optimize)
+            .set_search_paths(&search_paths)
             .set_frontend_opt(dialect > 21);
         let mut symbol_table = HashMap::new();
 
@@ -859,13 +928,13 @@ pub fn launch_tool(
 
         match res {
             Ok(r) => {
-                println!("{}", r.to_string());
+                stdout.write_string(r.to_string());
 
                 build_symbol_table_mut(&mut symbol_table, &r);
-                write_sym_output(&symbol_table, &"main.sym".to_string());
+                write_sym_output(&symbol_table, &"main.sym".to_string()).expect("writing symbols");
             }
             Err(c) => {
-                println!("{}: {}", c.0.to_string(), c.1);
+                stdout.write_string(format!("{}: {}", c.0.to_string(), c.1));
             }
         }
 
@@ -1017,7 +1086,6 @@ pub fn launch_tool(
             run_script,
             input_sexp.unwrap(),
             Some(RunProgramOption {
-                operator_lookup: None,
                 max_cost: if max_cost == 0 {
                     None
                 } else {
