@@ -3,10 +3,12 @@ use std::hash::{Hash, Hasher};
 use std::rc::Rc;
 use std::string::String;
 
+use binascii::{bin2hex, hex2bin};
 use num_traits::{zero, Num};
 
 use crate::classic::clvm::__type_compatibility__::{bi_zero, Bytes, BytesFromType};
 use crate::classic::clvm::casts::{bigint_from_bytes, bigint_to_bytes, TConvertOption};
+use crate::compiler::prims::prims;
 use crate::compiler::srcloc::Srcloc;
 use crate::util::{number_from_u8, u8_from_number, Number};
 
@@ -120,14 +122,40 @@ fn normalize_int(v: Vec<u8>, base: u32) -> Number {
     Number::from_str_radix(&s, base).unwrap()
 }
 
+// Hex values are _not_ numbers, they're byte strings expressed in hexadecimal
+// while they correspond numerically, integral constants are byte-padded to the
+// left to retain their sign and hex constants are considered unsigned so _not_
+// padded.
+fn from_hex(l: Srcloc, v: &Vec<u8>) -> SExp {
+    let mut result = vec![0; (v.len() - 2) / 2];
+    hex2bin(&v[2..], &mut result).expect("should convert from hex");
+    SExp::QuotedString(l, b'"', result)
+}
+
 fn make_atom(l: Srcloc, v: Vec<u8>) -> SExp {
     let alen = v.len();
     if alen > 1 && v[0] == b'#' {
+        // Search prims for appropriate primitive
+        let want_name = v[1..].to_vec();
+        for p in prims() {
+            if want_name == p.0 {
+                return p.1.clone();
+            }
+        }
+
+        // Fallback (probably)
         SExp::Atom(l, v[1..].to_vec())
     } else {
         match matches_integral(&v) {
-            Integral::Hex => SExp::Integer(l, normalize_int(v[2..].to_vec(), 16)),
-            Integral::Decimal => SExp::Integer(l, normalize_int(v, 10)),
+            Integral::Hex => from_hex(l, &v),
+            Integral::Decimal => {
+                let intval = normalize_int(v, 10);
+                if intval == bi_zero() {
+                    SExp::Nil(l)
+                } else {
+                    SExp::Integer(l, intval)
+                }
+            }
             Integral::NotIntegral => SExp::Atom(l, v),
         }
     }
@@ -231,7 +259,20 @@ impl SExp {
             SExp::Nil(_) => "()".to_string(),
             SExp::Cons(_, a, b) => format!("({})", list_no_parens(a, b)),
             SExp::Integer(_, v) => v.to_string(),
-            SExp::QuotedString(_, q, s) => format!("\"{}\"", escape_quote(*q, s)),
+            SExp::QuotedString(_, q, s) => {
+                if printable(s) {
+                    format!("\"{}\"", escape_quote(*q, s))
+                } else {
+                    let vlen = s.len() * 2;
+                    let mut outbuf = vec![0; vlen];
+                    bin2hex(s, &mut outbuf)
+                        .expect("should be able to convert unprintable string to hex");
+                    format!(
+                        "0x{}",
+                        std::str::from_utf8(&outbuf).expect("only hex digits expected")
+                    )
+                }
+            }
             SExp::Atom(l, a) => {
                 if a.is_empty() {
                     "()".to_string()
