@@ -28,7 +28,9 @@ fn run_string_maybe_opt(
     let runner = Rc::new(DefaultProgramRunner::new());
     let mut opts: Rc<dyn CompilerOpts> = Rc::new(DefaultCompilerOpts::new(&"*test*".to_string()));
     let srcloc = Srcloc::start(&"*test*".to_string());
-    opts = opts.set_frontend_opt(fe_opt);
+    opts = opts
+        .set_frontend_opt(fe_opt)
+        .set_search_paths(&vec!["resources/tests".to_string()]);
     let sexp_args = parse_sexp(srcloc.clone(), &args).map_err(|e| CompileErr(e.0, e.1))?[0].clone();
 
     compile_file(
@@ -248,7 +250,7 @@ fn run_test_7_maybe_opt(opt: bool) {
         ).unwrap();
     assert_eq!(
         result.to_string(),
-        "((51 305419896 1000000000))".to_string()
+        "((51 0x12345678 1000000000))".to_string()
     );
 }
 
@@ -547,8 +549,7 @@ fn test_defconstant() {
 
     assert_eq!(
         result.to_string(),
-        "((51 43124150325653191095732712509762329830013206679743532022320461771503765780085 2))"
-            .to_string()
+        "((51 0x5f5767744f91c1c326d927a63d9b34fa7035c10e3eb838c44e3afe127c1b7675 2))".to_string()
     );
 }
 
@@ -715,13 +716,313 @@ fn test_collatz_maybe_opt(opt: bool) {
         opt,
     )
     .unwrap();
-    assert_eq!(result.to_string(), "(q . 2)");
+    assert_eq!(result.to_string(), "2");
 }
 
+#[test]
 fn test_collatz() {
     test_collatz_maybe_opt(false);
 }
 
-fn test_collatz_opt() {
-    test_collatz_maybe_opt(true);
+#[test]
+fn fancy_nested_let_bindings_should_work() {
+    let result = run_string_maybe_opt(
+        &indoc! {"
+    (mod X
+     (include *standard-cl-21*)
+     (defun do-something (solutions se)
+        (if solutions
+             (let (
+                    (R (f solutions))
+                    (S (- se 99))
+                  )
+                (if (= (f solutions) 99)
+                    S
+                    (let ((something-else (+ se R)))
+                        (do-something (r solutions) something-else)
+                    )
+                )
+             )
+             100
+        )
+    )
+
+    (do-something X 1)
+    )
+        "}
+        .to_string(),
+        &"(1 2 3 100 99)".to_string(),
+        false,
+    )
+    .unwrap();
+    assert_eq!(result.to_string(), "8");
+}
+
+#[test]
+fn let_as_argument() {
+    let result = run_string_maybe_opt(
+        &indoc! {"
+    (mod (X)
+     (include *standard-cl-21*)
+     (defun twice (x) (* x 2))
+     (defun plus (x y) (+ x y))
+     (plus (let ((t (twice X))) t) 3))
+        "}
+        .to_string(),
+        &"(5)".to_string(),
+        false,
+    )
+    .unwrap();
+    assert_eq!(result.to_string(), "13");
+}
+
+#[test]
+fn recursive_let_complicated_arguments() {
+    let result = run_string_maybe_opt(
+        &indoc! {"
+    (mod (X Y)
+     (include *standard-cl-21*)
+     (defun G (A (@ pt (X Y))) (+ A X Y))
+     (defun F (@ pt (X Y)) (let ((X1 (+ X 1)) (Y1 (+ Y 3))) (G X1 (let ((p (list X1 Y1))) p))))
+     (F X Y)
+     )
+        "}
+        .to_string(),
+        &"(7 13)".to_string(),
+        false,
+    )
+    .unwrap();
+    assert_eq!(result.to_string(), "32");
+}
+
+#[test]
+fn test_let_structure_access_1() {
+    let result = run_string_maybe_opt(
+        &indoc! {"
+    (mod (X Y)
+      (include *standard-cl-21*)
+      (let ((a 1)
+            (b 2))
+        (let ((A (+ a 1))
+               (XX (+ X 1))
+               (C (+ b b)))
+         (if Y
+            (let ((D (+ C 1))
+                  (E (+ XX Y)))
+              (c D E)
+              )
+            (let ((D XX)
+                  (E (+ XX Y)))
+              (c D E)
+              )
+          )
+         )
+       )
+      )
+        "}
+        .to_string(),
+        &"(7 13)".to_string(),
+        false,
+    )
+    .unwrap();
+    // a = 1
+    // b = 2
+    // A = 2
+    // XX = X + 1
+    // C = 2 + 2
+    // if Y
+    //   D = 5
+    //   E = X + Y + 1
+    // else
+    //   D = X + 1
+    //   E = X + Y + 1
+    assert_eq!(result.to_string(), "(5 . 21)");
+}
+
+#[test]
+fn test_let_structure_access_2() {
+    let result = run_string_maybe_opt(
+        &indoc! {"
+    (mod (X Y)
+      (include *standard-cl-21*)
+      (let ((a 1)
+            (b 2))
+        (let* ((A (+ a 1))
+               (XX (+ X 1))
+               (C (+ A b)))
+         (if Y
+            (let ((D (+ C 1))
+                  (E (+ XX Y)))
+              (c D E)
+              )
+            (let ((D XX)
+                  (E (+ XX Y)))
+              (c D E)
+              )
+          )
+         )
+       )
+      )
+        "}
+        .to_string(),
+        &"(7 13)".to_string(),
+        false,
+    )
+    .unwrap();
+    // a = 1
+    // b = 2
+    // A = 2
+    // XX = X + 1
+    // C = 2 + 2
+    // if Y
+    //   D = 5
+    //   E = X + Y + 1
+    // else
+    //   D = X + 1
+    //   E = X + Y + 1
+    assert_eq!(result.to_string(), "(5 . 21)");
+}
+
+#[test]
+fn test_let_inline_1() {
+    let result = run_string_maybe_opt(
+        &indoc! {"
+    (mod (G)
+      (include *standard-cl-21*)
+      (defun-inline F (X) (let ((Y (* X 2))) (+ Y 1)))
+      (F G)
+      )
+        "}
+        .to_string(),
+        &"(5)".to_string(),
+        false,
+    )
+    .unwrap();
+    assert_eq!(result.to_string(), "11");
+}
+
+#[test]
+fn read_of_hex_constant_in_modern_chialisp() {
+    let result = run_string(
+        &indoc! {"(mod () (include *standard-cl-21*) (sha256 (q . 1) (q . 0xf22ada22a0ed015000ea157013ee62dc6ce337a649ec01054fc62ed6caac7eaf)))"}
+        .to_string(),
+        &"()".to_string(),
+    )
+        .unwrap();
+
+    assert_eq!(
+        result.to_string(),
+        "36364122602940516403027890844760998025315693007634105146514094828976803085567".to_string()
+    );
+}
+
+#[test]
+fn hash_handling_test_2() {
+    let result = run_string(
+        &indoc! {"(mod () (include *standard-cl-21*) (sha256 (q . 1) (q . 1)))"}.to_string(),
+        &"()".to_string(),
+    )
+    .unwrap();
+
+    assert_eq!(
+        result.to_string(),
+        "-44412188149083219915772186748035909266791016930429887947443501395007119841358"
+    );
+}
+
+#[test]
+fn hash_handling_test_3() {
+    let result = run_string(
+        &indoc! {"(mod () (include *standard-cl-21*) (sha256 1 0))"}.to_string(),
+        &"()".to_string(),
+    )
+    .unwrap();
+
+    assert_eq!(
+        result.to_string(),
+        "34356466678672179216206944866734405838331831190171667647615530531663699592602"
+    );
+}
+
+#[test]
+fn sebastian_hash_test_1() {
+    let result = run_string(
+        &indoc! {"
+(mod (MOD_HASH TOKEN_A_AMOUNT TOKEN_B_AMOUNT K token_a_delta token_b_delta)
+    (include \"condition_codes.clvm\")
+    (include \"curry-and-treehash.clinc\")
+    (include *standard-cl-21*)
+
+    (defun sha256tree1 (TREE)
+        (if (l TREE)
+            (sha256 2 (sha256tree1 (f TREE)) (sha256tree1 (r TREE)))
+            (sha256 1 TREE)
+        )
+    )
+    (defun return-new-coin (mod_hash new_token_a_amount new_token_b_amount K)
+        (list (list mod_hash (sha256tree1 mod_hash))
+            (list
+                CREATE_COIN
+               (puzzle-hash-of-curried-function
+                    mod_hash (sha256tree1 mod_hash)
+                )
+                1
+            )
+        )
+    )
+
+    (let (
+         (new_token_a_amount (+ TOKEN_A_AMOUNT token_a_delta))
+         (new_token_b_amount (+ TOKEN_B_AMOUNT token_b_delta))
+        )
+        (if (all (= K (* new_token_a_amount new_token_b_amount)) (> new_token_a_amount 0) (> new_token_b_amount 0) )
+            (return-new-coin MOD_HASH new_token_a_amount new_token_b_amount K)
+            (x new_token_a_amount new_token_b_amount)
+        )
+    ))"}
+        .to_string(),
+        &"(0xf22ada22a0ed015000ea157013ee62dc6ce337a649ec01054fc62ed6caac7eaf 10000 200 2000000 -2000 50)".to_string()
+    )
+        .unwrap();
+
+    assert_eq!(result.to_string(), "((0xf22ada22a0ed015000ea157013ee62dc6ce337a649ec01054fc62ed6caac7eaf 36364122602940516403027890844760998025315693007634105146514094828976803085567) (51 -54330418644829767769717664495663952010367061369724157609947295940464695774007 1))");
+}
+
+#[test]
+fn sebastian_hash_test_2() {
+    let result = run_string(
+        &indoc! {"
+(mod (MOD_HASH TOKEN_A_AMOUNT TOKEN_B_AMOUNT K token_a_delta token_b_delta)
+    (include \"condition_codes.clvm\")
+    (include \"curry-and-treehash.clinc\")
+
+    (defun sha256tree1 (TREE)
+        (if (l TREE)
+            (sha256 2 (sha256tree1 (f TREE)) (sha256tree1 (r TREE)))
+            (sha256 1 TREE)
+        )
+    )
+    (defun return-new-coin (mod_hash new_token_a_amount new_token_b_amount K)
+        (list (list mod_hash (sha256tree1 mod_hash))
+            (list
+                CREATE_COIN
+               (puzzle-hash-of-curried-function
+                    mod_hash (sha256tree1 mod_hash)
+                )
+                1
+            )
+        )
+    )
+
+        (if (all (= K (* (+ TOKEN_A_AMOUNT token_a_delta) (+ TOKEN_B_AMOUNT token_b_delta))) (> (+ TOKEN_A_AMOUNT token_a_delta) 0) (> (+ TOKEN_B_AMOUNT token_b_delta) 0) )
+            (return-new-coin MOD_HASH (+ TOKEN_A_AMOUNT token_a_delta) (+ TOKEN_B_AMOUNT token_b_delta) K)
+            (x (+ TOKEN_A_AMOUNT token_a_delta) (+ TOKEN_B_AMOUNT token_b_delta))
+        )
+    )"}
+        .to_string(),
+        &"(0xf22ada22a0ed015000ea157013ee62dc6ce337a649ec01054fc62ed6caac7eaf 10000 200 2000000 -2000 50)".to_string()
+    )
+        .unwrap();
+
+    assert_eq!(result.to_string(), "((0xf22ada22a0ed015000ea157013ee62dc6ce337a649ec01054fc62ed6caac7eaf 36364122602940516403027890844760998025315693007634105146514094828976803085567) (51 -54330418644829767769717664495663952010367061369724157609947295940464695774007 1))");
 }
